# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-07-30 19:57
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_light_enums.db
import magic_cards.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Artist',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('full_name', models.CharField(max_length=127, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='Card',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
                ('mana_cost', models.CharField(blank=True, max_length=63)),
                ('text', models.TextField(blank=True)),
                ('power', models.CharField(blank=True, max_length=7)),
                ('toughness', models.CharField(blank=True, max_length=7)),
            ],
            bases=(magic_cards.models.NameMixin, models.Model),
        ),
        migrations.CreateModel(
            name='CardSubtype',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=32, unique=True)),
            ],
            bases=(magic_cards.models.NameMixin, models.Model),
        ),
        migrations.CreateModel(
            name='CardSupertype',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=32, unique=True)),
            ],
            bases=(magic_cards.models.NameMixin, models.Model),
        ),
        migrations.CreateModel(
            name='CardType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=32, unique=True)),
            ],
            bases=(magic_cards.models.NameMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Printing',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('rarity', django_light_enums.db.EnumField(choices=[(40, b'COMMON'), (10, b'MYTHIC'), (50, b'SPECIAL'), (20, b'RARE'), (60, b'BASIC_LAND'), (30, b'UNCOMMON')], default=10, enum_values=[40, 10, 50, 20, 60, 30])),
                ('flavor_text', models.TextField(blank=True)),
                ('number', models.CharField(blank=True, max_length=7)),
                ('multiverse_id', models.PositiveIntegerField(blank=True, null=True)),
                ('artist', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='printings', to='magic_cards.Artist')),
                ('card', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='printings', to='magic_cards.Card')),
            ],
        ),
        migrations.CreateModel(
            name='Set',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=63, unique=True)),
                ('code', models.CharField(max_length=8, unique=True)),
            ],
            bases=(magic_cards.models.NameMixin, models.Model),
        ),
        migrations.AddField(
            model_name='printing',
            name='set',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='printings', to='magic_cards.Set'),
        ),
        migrations.AddField(
            model_name='card',
            name='subtypes',
            field=models.ManyToManyField(to='magic_cards.CardSubtype'),
        ),
        migrations.AddField(
            model_name='card',
            name='supertypes',
            field=models.ManyToManyField(to='magic_cards.CardSupertype'),
        ),
        migrations.AddField(
            model_name='card',
            name='types',
            field=models.ManyToManyField(to='magic_cards.CardType'),
        ),
    ]



from __future__ import unicode_literals

import random

from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django_light_enums import enum


@python_2_unicode_compatible
class NameMixin(object):
    def __str__(self):
        return self.name


class Card(NameMixin, models.Model):
    name = models.CharField(max_length=255, unique=True)
    mana_cost = models.CharField(max_length=63, blank=True)

    supertypes = models.ManyToManyField('CardSupertype')
    types = models.ManyToManyField('CardType')
    subtypes = models.ManyToManyField('CardSubtype')

    text = models.TextField(blank=True)
    power = models.CharField(max_length=7, blank=True)
    toughness = models.CharField(max_length=7, blank=True)
    loyalty = models.SmallIntegerField(blank=True, null=True)


class Set(NameMixin, models.Model):
    name = models.CharField(max_length=63, unique=True)
    code = models.CharField(max_length=8, unique=True)


class PrintingQuerySet(models.QuerySet):
    def random(self, num):
        num = int(num)
        printing_ids = set(self.values_list('id', flat=True))
        random_ids = random.sample(printing_ids, num)
        return self.filter(id__in=random_ids)


@python_2_unicode_compatible
class Printing(models.Model):
    class Rarity(enum.Enum):
        MYTHIC = 10
        RARE = 20
        UNCOMMON = 30
        COMMON = 40
        SPECIAL = 50
        BASIC_LAND = 60

    objects = PrintingQuerySet.as_manager()

    card = models.ForeignKey('Card', related_name='printings')
    set = models.ForeignKey('Set', related_name='printings')
    rarity = enum.EnumField(Rarity)
    flavor_text = models.TextField(blank=True)
    artist = models.ForeignKey('Artist', related_name='printings')
    number = models.CharField(max_length=7, blank=True)
    multiverse_id = models.PositiveIntegerField(blank=True, null=True)

    @property
    def image_url(self):
        if self.multiverse_id:
            return 'http://gatherer.wizards.com/Handlers/Image.ashx?multiverseid={}&type=card'.format(
                self.multiverse_id)

    def __str__(self):
        return '{} ({})'.format(self.card, self.set.code)


class CardSupertype(NameMixin, models.Model):
    name = models.CharField(max_length=32, unique=True)


class CardType(NameMixin, models.Model):
    name = models.CharField(max_length=32, unique=True)


class CardSubtype(NameMixin, models.Model):
    name = models.CharField(max_length=32, unique=True)


@python_2_unicode_compatible
class Artist(models.Model):
    full_name = models.CharField(max_length=127, unique=True)

    def __str__(self):
        return self.full_name



from django.core.management import BaseCommand
import inflect

from magic_cards.models import Card, Printing, Set
from magic_cards.utils.import_cards import import_cards, Everything


class Command(BaseCommand):
    help = 'Imports data from MTGJSON into your local database.'

    def add_arguments(self, parser):
        parser.add_argument('set_code', nargs='*', type=str)

    def handle(self, *args, **options):
        models_to_track = [Set, Card, Printing]
        initial = {model: model.objects.count() for model in models_to_track}

        p = inflect.engine()
        set_codes = options['set_code']
        if set_codes:
            count = len(set_codes)
            set_string = 'num({count}) plural_noun(set) ({codes})'.format(count=count, codes=', '.join(set_codes))
        else:
            set_string = 'all sets'

        self.stdout.write(p.inflect("Beginning import of {}.".format(set_string)))
        import_cards(set_codes or Everything)
        self.stdout.write("Import complete.")

        final = {model: model.objects.count() for model in models_to_track}
        status_strings = [
            p.inflect(
                "{0} new num({0},)plural_noun({1})".format(final[model] - initial[model], model._meta.object_name))
            for model in models_to_track
        ]
        self.stdout.write("Added {}.".format(p.join(status_strings)))



import random


def weighted_choice(choices):
    """
    Return a single element from a weighted sample.

    `choices` is a dictionary with labels (buckets) as keys and weights (probabilities) as values.
    """
    total = sum(choices.values())
    r = random.uniform(0, total)
    items = sorted(choices.items(), key=lambda x: x[1])
    for bucket, weight in items:
        r -= weight
        if r <= 0:
            return bucket



from django.contrib import admin

from .models import Card, Set, Printing, CardSupertype, CardType, CardSubtype, Artist


@admin.register(Card)
class CardAdmin(admin.ModelAdmin):
    search_fields = ['name']


@admin.register(Set)
class SetAdmin(admin.ModelAdmin):
    list_display = ['name', 'code']
    search_fields = ['name', 'code']


@admin.register(Printing)
class PrintingAdmin(admin.ModelAdmin):
    search_fields = ['card__name']
    list_filter = ['set']


@admin.register(Artist)
class ArtistAdmin(admin.ModelAdmin):
    search_fields = ['full_name']


@admin.register(CardSupertype)
class CardSupertypeAdmin(admin.ModelAdmin):
    search_fields = ['name']


@admin.register(CardType)
class CardTypeAdmin(admin.ModelAdmin):
    search_fields = ['name']


@admin.register(CardSubtype)
class CardSubtypeAdmin(admin.ModelAdmin):
    search_fields = ['name']



import io
import json
import zipfile
from contextlib import closing

import requests
from django.db import transaction

from magic_cards.models import Artist, Card, CardSubtype, CardSupertype, CardType, Printing, Set

MTG_JSON_URL = 'https://mtgjson.com/json/AllSets-x.json.zip'
FALLBACK_MTG_JSON_URL = 'http://mtgjson.com/json/AllSets-x.json.zip'


class Everything:
    """
    Sentinel value for downloading all sets (i.e. skipping nothing).
    """
    pass


def fetch_data():
    try:
        r = requests.get(MTG_JSON_URL)
    except requests.ConnectionError:
        r = requests.get(FALLBACK_MTG_JSON_URL)
    with closing(r), zipfile.ZipFile(io.BytesIO(r.content)) as archive:
        unzipped_files = archive.infolist()
        if len(unzipped_files) != 1:
            raise RuntimeError("Found an unexpected number of files in the MTGJSON archive.")
        data = archive.read(archive.infolist()[0])
    decoded_data = data.decode('utf-8')
    sets_data = json.loads(decoded_data)
    return sets_data


def parse_rarity(string):
    if string == 'Mythic Rare':
        return Printing.Rarity.MYTHIC
    elif string == 'Rare':
        return Printing.Rarity.RARE
    elif string == 'Uncommon':
        return Printing.Rarity.UNCOMMON
    elif string == 'Common':
        return Printing.Rarity.COMMON
    elif string == 'Basic Land':
        return Printing.Rarity.BASIC_LAND
    else:
        return Printing.Rarity.SPECIAL


class ModelCache(dict):
    def get_or_create(self, model, field, value, **kwargs):
        """
        Retrieves object of class `model` with lookup key `value` from the cache. If not found,
        creates the object based on `field=value` and any other `kwargs`.

        Returns a tuple of `(object, created)`, where `created` is a boolean specifying whether an
        `object` was created.
        """
        result = self[model].get(value)
        created = False
        if not result:
            kwargs[field] = value
            result = model.objects.create(**kwargs)
            self[model][value] = result
            created = True
        return result, created


def parse_data(sets_data, set_codes):
    # Load supertypes, types, and subtypes into memory
    cache = ModelCache()
    for model in [CardSupertype, CardType, CardSubtype]:
        cache[model] = {obj.name: obj for obj in model.objects.all()}
    # Load relevant sets into memory
    if set_codes is Everything:
        cache[Set] = {obj.code: obj for obj in Set.objects.all()}
    else:
        cache[Set] = {obj.code: obj for obj in Set.objects.filter(code__in=set_codes)}

    # Process the data set-by-set
    for code, data in sets_data.items():

        # Skip sets that have not been chosen
        if set_codes is not Everything and code not in set_codes:
            continue

        # Create the set
        magic_set, set_created = cache.get_or_create(Set, 'code', code, name=data['name'])

        printings_to_create = []

        # Create cards
        all_cards_data = data['cards']
        for card_data in all_cards_data:
            # Skip tokens
            layout = card_data['layout']
            if layout == 'token':
                continue

            # Card info
            name = card_data['name']
            mana_cost = card_data.get('manaCost', '')
            text = card_data.get('text', '')
            power = card_data.get('power', '')
            toughness = card_data.get('toughness', '')
            loyalty = card_data.get('loyalty', None)
            card, created = Card.objects.update_or_create(
                name=name, defaults={
                    'mana_cost': mana_cost,
                    'text': text,
                    'power': power,
                    'toughness': toughness,
                    'loyalty': loyalty,
                })
            supertypes = card_data.get('supertypes', [])
            types = card_data['types']
            subtypes = card_data.get('subtypes', [])
            if not created:
                card.supertypes.clear()
                card.types.clear()
                card.subtypes.clear()
            for supertype_name in supertypes:
                supertype, _ = cache.get_or_create(CardSupertype, 'name', supertype_name)
                card.supertypes.add(supertype)
            for type_name in types:
                card_type, _ = cache.get_or_create(CardType, 'name', type_name)
                card.types.add(card_type)
            for subtype_name in subtypes:
                subtype, _ = cache.get_or_create(CardSubtype, 'name', subtype_name)
                card.subtypes.add(subtype)

            # Printing info
            artist_name = card_data['artist']
            artist, _ = Artist.objects.get_or_create(full_name=artist_name)
            multiverse_id = card_data.get('multiverseid', None)  # Missing on certain sets
            flavor_text = card_data.get('flavor', '')
            rarity = card_data['rarity']
            number = card_data.get('number', '')  # Absent on old sets
            # If the Set was just created, we don't need to check if the Printing already exists,
            # and we can leverage bulk_create.
            printing_kwargs = {
                'card': card,
                'set': magic_set,
                'rarity': parse_rarity(rarity),
                'flavor_text': flavor_text,
                'artist': artist,
                'number': number,
                'multiverse_id': multiverse_id
            }
            if set_created:
                printings_to_create.append(Printing(**printing_kwargs))
            else:
                # Use .filter().exists() followed by a create() instead of get_or_create,
                # since these kwargs aren't unique for sets without proper multiverse_ids.
                if not Printing.objects.filter(**printing_kwargs).exists():
                    Printing.objects.create(**printing_kwargs)

        if printings_to_create:
            Printing.objects.bulk_create(printings_to_create)

    # Remove extra Printings caused by data that is duplicated on MTGJSON.
    # https://github.com/mtgjson/mtgjson/issues/388
    if set_codes is Everything or 'BOK' in set_codes:
        bugged_card_names = ['Jaraku the Interloper', 'Scarmaker']
        for name in bugged_card_names:
            extra_printings = Printing.objects.filter(
                set__code='BOK', card__name=name)[1:].values_list(
                    'pk', flat=True)
            Printing.objects.filter(pk__in=list(extra_printings)).delete()

    # Clean up any supertypes, subtypes, and types that have no Cards left.
    for model in [CardSubtype, CardType, CardSupertype]:
        for obj in model.objects.all():
            if obj.card_set.count() == 0:
                obj.delete()


@transaction.atomic
def import_cards(set_codes=Everything):
    sets_data = fetch_data()
    parse_data(sets_data, set_codes)


if __name__ == "__main__":
    import_cards()
